/*************************************************************************
 *                                                                       *
 * Copyright (C) 1993 Igor Mandrichenko.                                 *
 * Permission is granted to any individual or institution to use, copy,  *
 * or redistribute this software so long as all of the original files    *
 * are included unmodified, and that this copyright notice is retained.  *
 *                                                                       *
 *************************************************************************/

/*
 *    vms_pk.c  by Igor Mandrichenko
 *    version 2.0	20-mar-1993
 *                      Generates PKWARE version of VMS attributes
 *                      extra field according to appnote 2.0.
 *                      Uses low level QIO-ACP interface.
 *    version 2.0-1	10-apr-1993
 *			Save ACLs
 *    version 2.1	24-aug-1993
 *			By default produce 0x010C extra record ID instead of
 *			PKWARE's 0x000C. The format is mostly compatible with
 *			PKWARE.
 *			Incompatibility (?): zip produces multiple ACE
 *			fields.
 *    version 2.1-1	Clean extra fields in vms_get_attributes().
 *			Fixed bug with EOF.
 */

#include <descrip.h>
#include <rms.h>
#include <atrdef.h>
#include <fibdef.h>
#include <iodef.h>
#include <ssdef.h>
#include "zip.h"
#include "vmsmunch.h"

#define	VMS_ZIP
#include "vms.h"

#define ERR(x)	(((x)&1)==0)

#ifndef NULL
#define NULL (void*)(0L)
#endif

static struct PK_info PK_def_info = 
{
	ATR$C_RECATTR,	ATR$S_RECATTR,	{0},
	ATR$C_UCHAR,	ATR$S_UCHAR,	{0},
	ATR$C_JOURNAL,	ATR$S_JOURNAL,	{0},
	ATR$C_CREDATE,	ATR$S_CREDATE,	{0},
	ATR$C_REVDATE,	ATR$S_REVDATE,	{0},
	ATR$C_EXPDATE,	ATR$S_EXPDATE,	{0},
	ATR$C_BAKDATE,	ATR$S_BAKDATE,	{0},
	ATR$C_ASCDATES,	sizeof(ush), 	0,
	ATR$C_UIC,	ATR$S_UIC,	{0},
	ATR$C_FPRO,	ATR$S_FPRO,	{0},
	ATR$C_RPRO,	ATR$S_RPRO,	{0}
};

/*---------------*
 |  vms_open()   |
 *---------------*      
 |  This routine opens file for reading fetching its attributes.
 |  Returns pointer to file description structure.
 */

struct ioctx *vms_open(file,how)
char *file,*how;
{
    static struct atrdef	Atr[VMS_MAX_ATRCNT+1];
    static struct NAM		Nam;
    static struct fibdef	Fib;
    static struct dsc$descriptor FibDesc =
        {sizeof(Fib),DSC$K_DTYPE_Z,DSC$K_CLASS_S,&Fib};
    static struct dsc$descriptor_s DevDesc =
        {0,DSC$K_DTYPE_T,DSC$K_CLASS_S,&Nam.nam$t_dvi[1]};
    static struct dsc$descriptor_s FileName =
        {0,DSC$K_DTYPE_T,DSC$K_CLASS_S,0};
    static char EName[NAM$C_MAXRSS];
    static char RName[NAM$C_MAXRSS];

    struct FAB	Fab;
    register struct ioctx *ctx;
    register struct fatdef *fat;
    int status;
    int i;
    ulg efblk,hiblk;

    if( (ctx=(struct ioctx *)malloc(sizeof(struct ioctx))) == NULL )
	return NULL;     
    ctx -> PKi = PK_def_info;

#define FILL_REQ(ix,id,b)   {	    \
    Atr[ix].atr$l_addr = &(b);	    \
    Atr[ix].atr$w_type = (id);	    \
    Atr[ix].atr$w_size = sizeof(b); \
}
    FILL_REQ(0,	ATR$C_RECATTR,	ctx->PKi.ra);
    FILL_REQ(1,	ATR$C_UCHAR,	ctx->PKi.uc);
    FILL_REQ(2,	ATR$C_REVDATE,	ctx->PKi.rd);
    FILL_REQ(3,	ATR$C_EXPDATE,	ctx->PKi.ed);
    FILL_REQ(4,	ATR$C_CREDATE,	ctx->PKi.cd);
    FILL_REQ(5,	ATR$C_BAKDATE,	ctx->PKi.bd);
    FILL_REQ(6,	ATR$C_ASCDATES,	ctx->PKi.rn);
    FILL_REQ(7,	ATR$C_JOURNAL,	ctx->PKi.jr);
    FILL_REQ(8,	ATR$C_RPRO,	ctx->PKi.rp);
    FILL_REQ(9,	ATR$C_FPRO,	ctx->PKi.fp);
    FILL_REQ(10,ATR$C_UIC,	ctx->PKi.ui);
    FILL_REQ(11,ATR$C_ACLLENGTH,ctx->acllen);
    FILL_REQ(12,ATR$C_READACL,	ctx->aclbuf);

    Atr[13].atr$w_type = Atr[13].atr$w_size =	/* End of ATR list */
    Atr[13].atr$l_addr = (byte*)0L;

    /* initialize RMS structures, we need a NAM to retrieve the FID */
    Fab = cc$rms_fab;
    Fab.fab$l_fna = file ; /* name of file */
    Fab.fab$b_fns = strlen(file);
    Fab.fab$l_nam = &Nam; /* FAB has an associated NAM */
    Nam = cc$rms_nam;
    Nam.nam$l_esa = &EName; /* expanded filename */
    Nam.nam$b_ess = sizeof(EName);
    Nam.nam$l_rsa = &RName; /* resultant filename */
    Nam.nam$b_rss = sizeof(RName);
     
    /* do $PARSE and $SEARCH here */
    status = sys$parse(&Fab);
    if (!(status & 1)) return NULL;
     
    /* search for the first file.. If none signal error */
    status = sys$search(&Fab);
    if (!(status & 1)) return NULL;
     
    /* initialize Device name length, note that this points into the NAM
         to get the device name filled in by the $PARSE, $SEARCH services */
    DevDesc.dsc$w_length = Nam.nam$t_dvi[0];
     
    status = sys$assign(&DevDesc,&ctx->chan,0,0);
    if (!(status & 1)) return NULL;

    FileName.dsc$a_pointer = Nam.nam$l_name;
    FileName.dsc$w_length = Nam.nam$b_name+Nam.nam$b_type+Nam.nam$b_ver;

    /* Initialize the FIB */
    Fib.fib$r_acctl_overlay.fib$l_acctl = FIB$M_NOWRITE;
    for (i=0;i<3;i++)
        Fib.fib$r_fid_overlay.fib$w_fid[i]=Nam.nam$w_fid[i];
    for (i=0;i<3;i++)
	Fib.fib$r_did_overlay.fib$w_did[i]=Nam.nam$w_did[i];
     
    /* Use the IO$_ACCESS function to return info about the file */
    status = sys$qiow(0,ctx->chan,IO$_ACCESS|IO$M_ACCESS,&ctx->iosb,0,0,
		    &FibDesc,&FileName,0,0,&Atr,0);

    if (ERR(status) || ERR(status = ctx->iosb.status))
    {	vms_close(ctx);
	return NULL;
    }

    fat = (struct fatdef *)&(ctx -> PKi.ra);

#define SWAPW(x)	( (((x)>>16)&0xFFFF) + ((x)<<16) )

    efblk = SWAPW(fat->fat$r_efblk_overlay.fat$l_efblk);
    hiblk = SWAPW(fat->fat$r_hiblk_overlay.fat$l_hiblk);

    if( efblk == 0 && fat -> fat$w_ffbyte == 0 )
	ctx -> size =
	ctx -> rest = hiblk * 512;
    else
	ctx -> size =
	ctx -> rest = (efblk - 1) * 512 + fat -> fat$w_ffbyte;

    ctx -> status = SS$_NORMAL;
    ctx -> vbn = 1;
    return ctx;
}

#define KByte	1024

/*----------------*
 |   vms_read()   |
 *----------------*
 |   Reads file block by block into the buffer.
 |   Stops on EOF, returns number of bytes actualy read.
 |   Note: size of the buffer must be greater or equal to 512 !
 */

int vms_read(ctx, buf, size)
register struct ioctx *ctx;
register char *buf;
register int size;
{
    int status;
    long nr=0;

    if (ctx -> rest <= 0 || ctx -> status == SS$_ENDOFFILE)
        return 0;               /* Eof */

    if(size <= 0)
	return 0;

    if(size > 16*KByte)
        size = 16*KByte;
    else if(size > 512)
        size &= ~511L;	/* Round to integer number of blocks */

    do
    {   status = sys$qiow(0, ctx->chan, IO$_READVBLK,
	    &ctx->iosb, 0, 0,
	    buf, size, ctx->vbn,0,0,0);

	ctx->vbn += size>>9;
	if( size < 512 )
		++ctx->vbn;

        if( !ERR(status) )
		status = ctx->iosb.status;

        if( !ERR(status) || status == SS$_ENDOFFILE )
        {   register int count;

	    if( status == SS$_ENDOFFILE )
		count = ctx->rest;
	    else
		count = ctx->iosb.count;

	    size -= count;
	    buf  += count;
	    nr   += count;
	}
    } while( !ERR(status) && size > 0 );

    if (!ERR(status))
    {	ctx -> status = SS$_NORMAL;
	ctx -> rest -= nr;
	return nr;
    }
    else if(status == SS$_ENDOFFILE)
    {	ctx -> status = SS$_ENDOFFILE;
	ctx -> rest = 0;
	return nr;
    }
    else
    {	ctx -> status = status;
	ctx -> rest -= nr;
	return nr;
    }
}

/*-----------------*
 |   vms_error()   |
 *-----------------*
 |   Returns whether last operation on the file caused an error
 */

int vms_error(ctx)
struct ioctx *ctx;
{   /* EOF is not actual error */
    return ERR(ctx->status) && (ctx->status != SS$_ENDOFFILE);
}

/*------------------*
 |   vms_rewind()   |
 *------------------*
 |   Rewinds file to the beginning for the next vms_read().
 */

int vms_rewind(ctx)
struct ioctx *ctx;
{
    ctx -> vbn = 1;
    ctx -> rest = ctx -> size;
    return 0;
}

/*--------------------------*
 |   vms_get_attributes()   |
 *--------------------------*
 |   Returns pointer to malloc'ed PKWARE extra field with file attributes.
 |   File must be open already by vms_open()
 */

void vms_get_attributes(ctx,z)
struct ioctx *ctx;
struct zlist far *z;    /* zip entry to compress */
{	
	byte	*p,*b;
	struct	PK_header    *h;
	int l;

	/* Free extra fields restored from zipfile */
	if( z->ext )
	{   free(z->extra);
	    z->ext = 0;
	    if( (byte*)z->extra != (byte*)z->cextra && z->cext )
	    	free(z->cextra);
	    z->cext = 0;
	}

	if( z->cext )
	{   free(z->cextra);
	    z->cext = 0;
	}

	if( !vms_native )
	    return;
	
	l = PK_HEADER_SIZE+sizeof(ctx->PKi);
	if(ctx->acllen > 0)
	    l += PK_FLDHDR_SIZE + ctx->acllen;

	b = (uch *)malloc(l);
	if( b==NULL )
		return;

	h = (struct PK_header *)b;
	h -> tag = dosify ? PK_SIGNATURE : IZ_NEW_SIGNATURE;
	h -> size = l - PK_HEADER_SIZE;
	p = &(h->data);

	/* Copy default set of attributes */
	memcpy(&h->data,(uch*)&(ctx->PKi),sizeof(ctx->PKi));
	p += sizeof(ctx->PKi);

	if( ctx->acllen > 0 )
	{   struct PK_field *f;

	    if(dosify)
		warn("file has ACL, may be incompatible with PKUNZIP","");

	    f = (struct PK_field *)p;
	    f->tag = ATR$C_ADDACLENT;
	    f->size = ctx->acllen;
	    memcpy(&f->value[0],ctx->aclbuf,ctx->acllen);
	    p += PK_FLDHDR_SIZE + ctx->acllen;
	}


	updcrc(NULL,0);	    /* Init CRC register */
	h -> crc32 = updcrc(&h->data, h->size);

	z->ext = z->cext = l;
	z->extra = z->cextra = b;
}

void vms_close(ctx)
struct ioctx *ctx;
{
	sys$dassgn(ctx->chan);
      	free(ctx);
}

int vms_stat(file,s)
char *file;
stat_t *s;
{   int status;
    int staterr;
    struct FAB fab;
    struct XABFHC fhc;

    /*
    *  In simplest case when stat() returns "ok" and file size is
    *  nonzero or this is directory, finish with this
    */

    if( (staterr=stat(file,s)) == 0     
        && ( (int)(s->st_size) >= 0               /* Size - ok */
             || (s->st_mode & S_IFREG) == 0       /* Not a plain file */
           )
    ) return staterr;

    /*
    * Get here to handle the special case when stat() returns
    * invalid file size. Use RMS to compute the size.
    * When EOF block is zero, set file size to its physical size.
    * One more case to get here is when this is remote file accessed
    * via DECnet.
    */

    fab = cc$rms_fab;
    fhc = cc$rms_xabfhc;
    fab.fab$l_fna = file;
    fab.fab$b_fns = strlen(file);
    fab.fab$l_xab = (char*)(&fhc);

    fab.fab$b_fac = FAB$M_GET;
    
    status = sys$open(&fab);
    fab.fab$l_xab = (char*)0L;
    sys$close(&fab);

    if( !ERR(status) )
    {
        if( fhc.xab$l_ebk > 0 )
            s->st_size = ( fhc.xab$l_ebk-1 ) * 512 + fhc.xab$w_ffb;
        else if( fab.fab$b_org == FAB$C_IDX
                 || fab.fab$b_org == FAB$C_REL
                 || fab.fab$b_org == FAB$C_HSH )
		/* Special case, when ebk=0: save entire allocated space */
		    s->st_size = fhc.xab$l_hbk * 512;
	else
	    s->st_size = fhc.xab$w_ffb;
        return 0; /* stat() success code */
    }
    else
        return status;
}

